// Array Structure class

extern ofstream file;

// Since the implementation particulars are not salient to this
// assignment, we use a simple fixed-size array, with copying
// on insertion.
// Ideally, we'd use something more elegant and much faster for 
// the implementation, like a rotating buffer that allows
// constant-time insertion at either end, by holding a "head" and
// a "tail" index inside the class.
// If the size of the buffer is a power of two, wraparounds can
// be made very efficient.
// However, we here offer the abovementioned simpler solution.

// In addition, we could improve the amount of shared code slightly 
// by deriving privately from an ImplementationClass that held
// the size data member and corresponding inlined source code, 
// much as in Part 1.  However, that would be a change to the
// given hierarchy for the mixin, so we duplicated the data
// member and the code in this implementation.

template <class ImplementationDataType>
class Array_Structure
{
public:
  ////////////////////////////////////////////////////////////
  // Type definitions; renamed so as not to conflict with
  // the identical type definitions in Queue.
  //
  typedef ImplementationDataType (*ImplementationDequeueEmptyHandler)();

private:
  //////////////////////////////////////////////////////////////
  // State
  // 
	int max;
  int size;
	ImplementationDataType * array;


public:
  //////////////////////////////////////////////////////////////
  // Constructors
  // 
	Array_Structure() : max(0), array(0), size(0) {}
	Array_Structure(int maxSize) :
    max(maxSize), array(new ImplementationDataType [maxSize]), size(0) {}
	Array_Structure(const Array_Structure &source)
      { copyFrom(source); }
	// destructor frees array space
	~Array_Structure() {
    delete [] array;
  }

	// = operator
	Array_Structure & operator=(const Array_Structure &source) {
    if (&source!=this) {
      if (max<source.max) {
        // Resize this array.
        delete [] array;
        copyFrom(source);
      } else {
        copyDataFrom(source);
      }
    } else { file << "Warning: self-assignment\n"; }
    return *this;
  }


  //////////////////////////////////////////////////////////////
  // Behaviour methods
  // 

  int get_size() const { return size; }
	// includes searches for an element an returns true if it finds it
	bool includes(const ImplementationDataType & element) const {
    for (int i=0;i!=size;i++)
      if (array[i]==element)
        return true;
    return false;
  }
	// adds a new element in the array, at the end.
  void addLast (const ImplementationDataType & element) {
	  if (size < max) 
      { array[size++] = element; }
	  else 
      { file << "overflow ignoring " << element << endl; }
  }

	// adds a new element in the array, at the front.
	void addFirst (const ImplementationDataType & element) {
    if (size < max) 
      { insertAtFront(element); } 
    else
      { file << "overflow ignoring " << element << endl; }
  }

	// removeFirst 
	ImplementationDataType removeFirst(ImplementationDequeueEmptyHandler func_ptr) {
    return size==0 ? func_ptr() : removeFromFront();
  }

	// removeLast 
	ImplementationDataType removeLast(ImplementationDequeueEmptyHandler func_ptr) {
	  ImplementationDataType answer;

	  if (!size) { answer = func_ptr(); } 
	  else { answer = array[(size--) - 1]; }
	  
	  return answer;
  }

	// print outputs all elements in the array in the order they are stored
	void print() const {
    int i;

	  file << "ArrayStructure contents: <";
  
	  for (i = 0; i < size-1; i++)
  	  { file << array[i] << " "; }
	  file << array[i] << ">"<< endl;
  }

private:
  //////////////////////////////////////////////////////////////
  // Implementation methods
  // 
  void copyDataFrom(const Array_Structure &source) {
    for (int i=0;i!=max;i++)
      array[i]=source.array[i];
    size=source.size;
  }

  // Precondition:  array is not pointing at allocated memory.
  void copyFrom(const Array_Structure &source) {
    array=new ImplementationDataType [source.max];
    max=source.max;
    copyDataFrom(source);
  }

  // Precondition:  array is not full.
  void insertAtFront(const ImplementationDataType &element) {
    for (int i = size - 1; i >= 0; i--)
      { array[i+1] = array[i]; }

    array[0] = element; 
    size++;
  }

  // Precondition:  array is not empty.
  ImplementationDataType removeFromFront() {
		ImplementationDataType answer = array[0]; 
		for (int i = 1; i < size; i++)
      array[i - 1] = array[i];
    size--;
    return answer;
  }
};
